@@ -17,14 +17,31 @@
 	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
+#if defined(linux) || defined(__linux) || defined(__linux__)
+#undef __LINUX__
+#define __LINUX__	1
+#elif defined(__FreeBSD__) || defined(__DragonFly__) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__APPLE__)
+#undef __BSD__
+#define __BSD__		1
+#endif
+
 #include <stdlib.h>
 #include <stdio.h>
-#include <malloc.h>
 #include <sys/types.h>
 #include <string.h>
 #include <unistd.h>
 #include <stdarg.h>
 #include <assert.h>
+#include <errno.h>
+
+// used to check if device is mounted
+#if defined(__LINUX__)
+#include <mntent.h>
+#elif defined (__BSD__)
+#include <sys/param.h>
+#include <sys/ucred.h>
+#include <sys/mount.h>
+#endif
 
 // file attributes
 #define ATTR_READ_ONLY 0x01
@@ -54,13 +71,14 @@
 
 // program information
 #define INFO_PROGRAM "FATSort Utility"
-#define INFO_VERSION "0.9.6.1"
-#define INFO_AUTHOR "Boris Leidner <fatsort@formenos.de>"
+#define INFO_VERSION "0.9.7"
+#define INFO_AUTHOR "Boris Leidner <fatsort(at)formenos.de>"
 #define INFO_HEADER INFO_PROGRAM " " INFO_VERSION " by " INFO_AUTHOR
 #define INFO_USAGE "\n"	"Usage: fatsort [options] device\n" \
 			"\n" \
 			"Options:\n" \
 			"\t-c\t Ignore case of file names\n" \
+			"\t-r\t Sort in reverse order\n" \
 			"\t-h\t Print some help\n" \
 			"\t-i\t Print file system information only\n" \
 			"\t-l\t Print current order of files only\n" \
@@ -205,7 +223,7 @@ const struct sClusterChain __INITCLUSTERCHAIN__ = {0, NULL};
 
 // --------- option flags
 
-u_int32_t OPT_VERSION, OPT_HELP, OPT_INFO, OPT_QUIET, OPT_IGNORE_CASE, OPT_ORDER, OPT_LIST;
+u_int32_t OPT_VERSION, OPT_HELP, OPT_INFO, OPT_QUIET, OPT_IGNORE_CASE, OPT_ORDER, OPT_LIST, OPT_REVERSE;
 
 // ----------
 
@@ -394,9 +412,9 @@ int32_t cmpEntries(struct sDirEntryList *de1, struct sDirEntryList *de2) {
 	}
 
 	if (OPT_IGNORE_CASE) {
-		return strcasecmp(s1, s2);
+		return strcasecmp(s1, s2) * OPT_REVERSE;
 	} else {
-		return strcmp(s1, s2);
+		return strcmp(s1, s2) * OPT_REVERSE;
 	}
 }
 
@@ -920,12 +938,11 @@ void freeClusterChain(struct sClusterChain *chain) {
 
 	assert(chain != NULL);
 
-	struct sClusterChain *tmp;
+	struct sClusterChain *tmp, *next;
 	
-	while(chain->next != NULL) {
-		tmp=chain->next;
-		chain=chain->next;
-		free(tmp);
+	for (tmp = chain->next; tmp; tmp = next) {
+		next = tmp -> next;
+		free (tmp);
 	}
 }
 
@@ -1083,7 +1100,7 @@ int32_t sortClusterChain(FILE *fd, struct sBootSector *bs, u_int32_t cluster, ch
 
 	u_int32_t clen, value, c;
 	struct sClusterChain ClusterChain=__INITCLUSTERCHAIN__;
-	char newpath[MAX_PATH_LEN+1];
+	char newpath[MAX_PATH_LEN+1]={0};
 
 	struct sDirEntryList list = __INITDIRLIST__;
 	struct sDirEntryList *p;
@@ -1136,6 +1153,7 @@ int32_t sortClusterChain(FILE *fd, struct sBootSector *bs, u_int32_t cluster, ch
 			}
 
 			strncpy(newpath, path, MAX_PATH_LEN - strlen(newpath));
+			newpath[MAX_PATH_LEN]='\0';
 			if ((p->lname != NULL) && (p->lname[0] != '\0')) {
 				strncat(newpath, p->lname, MAX_PATH_LEN - strlen(newpath));
 				newpath[MAX_PATH_LEN]='\0';
@@ -1170,7 +1188,7 @@ int32_t sort_FAT16_rootdir(FILE *fd, struct sBootSector *bs) {
 
 	u_int32_t FATSz, BSOffset, i, ret;
 	u_int32_t entries,count=0, c, value;
-	char newpath[MAX_PATH_LEN+1];
+	char newpath[MAX_PATH_LEN+1]={0};
 	char empty[32]={0};
 
 	struct sDirEntryList list = __INITDIRLIST__;
@@ -1240,6 +1258,7 @@ int32_t sort_FAT16_rootdir(FILE *fd, struct sBootSector *bs) {
 			}
 
 			strncpy(newpath, "/", MAX_PATH_LEN - strlen(newpath));
+			newpath[MAX_PATH_LEN]='\0';
 			if ((p->lname != NULL) && (p->lname[0] != '\0')) {
 				strncat(newpath, p->lname, MAX_PATH_LEN - strlen(newpath));
 				newpath[MAX_PATH_LEN]='\0';
@@ -1277,7 +1296,7 @@ int32_t printFSInfo(char *filename) {
 
 	printf("\t- File system information -\n");
 
-	if ((fd=fopen(filename, "r+")) == NULL) {
+	if ((fd=fopen(filename, "r")) == NULL) {
 		perror(__func__);
 		return -1;
 	}
@@ -1332,6 +1351,57 @@ int32_t printFSInfo(char *filename) {
 
 }
 
+u_int32_t check_mounted(char *filename) {
+
+#if defined(__LINUX__)
+	FILE *fd;
+	struct mntent *mnt;
+	u_int32_t ret = 0;
+	
+	if ((fd = setmntent("/etc/mtab", "r")) == NULL) {
+		myerror((const char *) strerror(errno));
+		return 0;
+	}
+	
+	while ((mnt = getmntent(fd)) != NULL) {
+		if (strcmp(mnt->mnt_fsname, filename) == 0) {
+			ret = 1;
+			break;
+		}
+	}
+	
+	if (endmntent(fd) != 1) {
+		myerror("Closing mtab failed!");
+		return 0;
+	}
+
+	return ret;
+#elif defined(__BSD__)
+	struct statfs *mntbuf;
+	int i, mntsize;
+	u_int32_t ret = 0;
+
+	mntsize = getmntinfo(&mntbuf, MNT_NOWAIT);
+
+	if (mntsize == 0) {
+		myerror(strerror (errno));
+		return 0;
+	}
+	
+	for (i = mntsize - 1; i >= 0; i--) {
+		if (strcmp(mntbuf[i].f_mntfromname, filename) == 0) {
+			ret = 1;
+			break;
+		}
+	}
+	
+	return ret;
+#else
+	return 0;
+#endif
+}
+
+
 int32_t sort_fs(char *filename) {
 /*
 	sort FAT file system
@@ -1344,7 +1414,12 @@ int32_t sort_fs(char *filename) {
 
 	int32_t FATType;
 
-	if ((fd=fopen(filename, "r+")) == NULL) {
+	if (check_mounted (filename) && !OPT_LIST) {
+		myerror("Filesystem is mounted!");
+		return -1;
+	}
+
+	if ((fd=fopen(filename, (OPT_LIST) ? "r" : "r+")) == NULL) {
 		perror(__func__);
 		return -1;
 	}
@@ -1403,8 +1478,11 @@ int main(int argc, char *argv[]) {
 	printf("SizeofFAT32: %u\n", sizeof(struct sFAT32));
 	return 0;*/
 
+	/* Default (1) is normal order, use -1 for reverse order. */
+	OPT_REVERSE = 1;
+	
 	opterr=0;
-	while ((c=getopt(argc, argv, "ivhqco:l")) != -1) {
+	while ((c=getopt(argc, argv, "ivhqco:lr")) != -1) {
 		switch(c) {
 			case 'i' : OPT_INFO = 1; break;
 			case 'v' : OPT_VERSION = 1; break;
@@ -1412,6 +1490,7 @@ int main(int argc, char *argv[]) {
 			case 'q' : OPT_QUIET = 1; break;
 			case 'c' : OPT_IGNORE_CASE = 1; break;
 			case 'l' : OPT_LIST = 1; break;
+			case 'r' : OPT_REVERSE = -1; break;
 			case 'o' :
 				switch(optarg[0]) {
 					case 'd': OPT_ORDER=0; break;
